#*******************************************************************************
# (c) Copyright Charles Kooperberg 1993, 1995                                  *
# This function is part of Hazard Regression.                                  *
# It was written by Charles Kooperberg at the University of Washington         *
# in 1992 and 1993. It is described in the document:                           *
# Kooperberg, C, Stone, C. J. and Truong, Y. K. (1995) ``Hazard Regression'',  *
# Journal of the American Statistical Association, 90, 78-94.                  *
# You are free to use this program, for non-commercial purposes only,          *
# under two conditions:                                                        *
# (1) This note is not to be removed                                           *
# (2) Publications using HARE  computations should reference the publication   *
#     mentioned above.                                                         *
# For questions, please email clk@stat.washington.edu                          *
#                       Charles Kooperberg, November 29, 1995.                 *
#******************************************************************************/
hare.fit <- function(data, delta, cov, penalty, maxdim, exclude,
	include, prophaz = F, additive = F, linear, fit, silent = T)
{
  # get the parameters from the C-program
  call <- match.call()
  MAXKNOTS <- -3
  MAXSPACE <- -3
  z <- .C("share",
          mk = as.integer(MAXKNOTS),
          ms = as.integer(MAXSPACE))
  MAXKNOTS <- z$mk
  MAXSPACE <- z$ms
  # 
  # a few elementary data checks
  if(missing(data)) stop("there has to be data")
  if(length(data) < 25) stop("not enough data")
  if(min(data) < 0) stop("negative data")
  if(missing(delta))
    delta <- data - data + 1
  if(length(data) != length(delta))
    stop("data and delta have different length")
  dd <- abs(delta - 0.5)
  if(min(dd) < 0.5 || max(dd) > 0.5) stop("delta not all 0 or 1")
  ndata <- length(data)
  #
  # dealing with the covariates, sorting the cases 
  # first if there are no covariates
  if(missing(cov)) {
    ncov <- 0
    cov <- 0
    iia <- order(data)
    delta <- delta[iia]
    data <- data[iia]
  } else {
    if(length(cov) == ndata)
      cov <- matrix(cov, ncol = 1, nrow = ndata)
    if(length(cov[, 1]) != ndata)
      stop("covariates not ndata * ncov matrix")
    ncov <- length(cov[1,  ])
    cov <- cbind(cov, 1)
    y <- cbind(data, cov)
    keys <- 1:ndata
    for(i in (ncov + 2):1)
      keys <- keys[sort.list(y[keys, i])]
    data <- data[keys]
    delta <- delta[keys]
    cov <- cov[keys, 1:ncov]
  }
  if(additive) {
    if(!missing(exclude)) stop("cannot have exclude and additive")
    if(!missing(include)) stop("cannot have include and additive")
    prophaz <- F
    include <- c(0, 0)
  }
  if(missing(exclude) + missing(include) == 0)
    stop("only 1 from exclude and include allowed")
  vexclude <- 0                         #
  # using exclude
  if(missing(exclude) == F) {
    if(length(exclude) == 2)
      exclude <- matrix(exclude, ncol = 2, nrow = 1)
    if(length(exclude[1,  ]) != 2)
      stop("exclude has wrong shape")
    if(min(exclude) < 0 || max(exclude) > ncov)
      stop("exclude has wrong values")
    vexclude <- as.vector(t(exclude))
    vexclude <- c(length(vexclude)/2, vexclude)#
    # proportional hazards model
    if(prophaz && ncov > 0) {
      vexclude <- c(vexclude, as.vector(rbind(1:ncov, 0)))
      vexclude[1] <- vexclude[1] + ncov
    }
  }
  #
  # using include
  if(missing(include) == F || additive) {
    if(length(include) == 2)
      include <- matrix(include, ncol = 2, nrow = 1)
    if(length(include[1,  ]) != 2)
      stop("include has wrong shape")
    if(min(include) < 0 || max(include) > ncov)
      stop("include has wrong values")
    include <- t(apply(include, 1, sort))
    if(length(include) == 2)
      include <- matrix(include, ncol = 2, nrow = 1)
    if(prophaz)
      include <- include[include[, 1] > 0,  ]
    vexclude <- as.vector(t(include))
    vexclude <- c( - length(vexclude)/2, vexclude)
  }
  #
  # using proprtional hazards
  if(missing(include) && missing(exclude) && prophaz && ncov > 0) 
    vexclude <- c(ncov, as.vector(rbind(1:ncov, 0)))	
  # set parameters
  mindist <- 5
  if(missing(penalty))
    penalty <- log(ndata)
  if(missing(maxdim)) {
    maxdim <- floor(6 * (ndata)^0.2)
    if(maxdim > MAXSPACE - 1) maxdim <- MAXSPACE - 1
    maxdim <-  - maxdim
  }
  if(maxdim > MAXSPACE - 1) {
    maxdim <- MAXSPACE - 1
    print(paste("maximum dimension reduced to", maxdim))
  }
  lins <- rep(0, MAXSPACE)
  if(!missing(linear)) {
    linear[linear <= 0] <- ncov + 1
    linear[linear > ncov + 1] <- ncov + 1
    lins[linear] <- 1
  }
  if(additive) vexclude <- c(-1, 0, 0)
  # do it
  fitter <- 0
  bbtt <- matrix(0, ncol = 6, nrow = abs(maxdim))
  cckk <- matrix(0, ncol = (MAXKNOTS + 1), nrow = (ncov + 1))
  if(!missing(fit)) {
    fitter <- fit$ndim
    if(fit$ncov != ncov) stop("ncov and fit's ncov are different")
    bbtt[1:fit$ndim,  ] <- fit$fcts
    bbtt <- as.vector(t(bbtt))
    bbtt[is.na(bbtt)] <- -1
    a1 <- length(fit$knots[1,  ])
    cckk[, 1:a1] <- fit$knots
    cckk <- as.vector(cckk)
    cckk[is.na(cckk)] <- -1
  }
  z <- .C("share",
          as.integer(ncov),
          ndim = as.integer(ndata),
          as.double(data),
          as.integer(delta),
          as.double(cov),
          as.double(penalty),
          as.integer(mindist),
          as.integer(maxdim),
          bbtt = as.double(bbtt),
          cckk = as.double(cckk),
          as.integer(vexclude),
          as.integer(lins),
          as.integer(silent),
          logl = as.double(rep(0, MAXSPACE)),
          as.integer(fitter),
          ad = as.integer(rep(0, MAXSPACE)),
          as.integer(0))                # 
  # organize bbtt and cckk
  maxdim <- abs(maxdim)
  z$bbtt <- matrix(z$bbtt, nrow = maxdim, ncol = 6, byrow = T)[1:z$ndim,]
  z$cckk <- matrix(z$cckk, nrow = ncov + 1, ncol = MAXKNOTS + 1, byrow = T)
  z$cckk <- z$cckk[, 1:(1 + max(z$cckk[, 1]))]
  z$cckk <- matrix(z$cckk, nrow = ncov + 1)
  l1 <- max(z$cckk[, 1])
  for(i in 1:(ncov + 1))
    if(z$cckk[i, 1] != l1) z$cckk[i, (z$cckk[i, 1] + 2):(l1 + 1)] <- NA
  if(l1 > 0 && ncov > 0)
    dimnames(z$cckk) <- list(c("T", 1:ncov), c("K", 1:l1))
  if(l1 > 0 && ncov == 0)
    dimnames(z$cckk) <- list(c("T"), c("K", 1:l1))
  if(l1 == 0 && ncov > 0)
    dimnames(z$cckk) <- list(c("T", 1:ncov), "K")
  if(l1 == 0 && ncov == 0)
    dimnames(z$cckk) <- list(c("T"), "K")
  l1 <- max((1:MAXSPACE)[z$logl > -1e+100])
  z$bbtt <- matrix(z$bbtt, ncol = 6)
  dimnames(z$bbtt) <- list(1:(z$ndim), c("dim1", "knot1", "dim2", "knot2",
                                         "beta", "SE"))
  z$bbtt[z$bbtt[, 3] == -1, 3:4] <- NA
  if(is.na(l1)) {
    z$logl <- z$logl[1]
    z$ad <- z$ad[1]
  }
  else {
    z$logl <- z$logl[1:l1]
    z$ad <- z$ad[1:l1]
  }
  z$ad[z$logl < -1e+100] <- NA
  z$logl[z$logl < -1e+100] <- NA
  z$logl <- cbind(z$logl, z$ad)
  dimnames(z$logl) <- list(NULL, c("log-lik", "A/D"))
  ranges <- NA
  if(ncov == 1) ranges <- matrix(range(cov), ncol = 1, nrow = 2)
  if(ncov > 1) ranges <- apply(cov, 2, range)
  # done
  list(call = call, ncov = ncov, ndim = z$ndim, fcts = z$bbtt, knots 
       = z$cckk, penalty = penalty, max = max(data), ranges = ranges,
       logl = z$logl, sample = ndata)
}
hare.plot <- function(fit, cov, n = 100, which = 0, what = "d", time, add = F,
	xlim, xlab, ylab, type, ...)
{
  nocov <- 0
  if(fit$ncov == 0) nocov <- 1
  else if(length(cov) != fit$ncov) stop("covariates are wrong")
  if(which == 0) {
    if(missing(xlim)) {
      if(nocov == 0) {
        u1 <- qhare(0.01, cov, fit)
        u2 <- qhare(0.99, cov, fit)
      } else {
        u1 <- qhare(0.01, fit = fit)
        u2 <- qhare(0.99, fit = fit)
      }
      u3 <- 1.1 * u1 - 0.1 * u2
      u2 <- min(u2, fit$max)
      u4 <- 1.1 * u2 - 0.1 * u1
      if(u3 < 0) u3 <- 0
      else if(u4/u3 > 5) u3 <- 0
    } else {
      u3 <- xlim[1]
      u4 <- xlim[2]
    }
    xx <- (0:(n - 1))/(n - 1) * (u4 - u3) + u3
    if(fit$ncov > 0) yy <- cov
  } else {
    if(which < 0 || which > fit$ncov) stop("which is wrong")
    if(missing(time)) stop("time is missing")
    if(missing(xlim)) {
      u3 <- fit$ranges[1, which]
      u4 <- fit$ranges[2, which]
    } else {
      u3 <- xlim[1]
      u4 <- xlim[2]
    }
    xx <- (0:(n - 1))/(n - 1) * (u4 - u3) + u3
    yy <- matrix(cov, ncol = fit$ncov, nrow = n, byrow = T)
    yy[, which] <- xx
    xx <- time
  }
  iwhat <- 0
  if(what == "d" || what == "D") iwhat <- 3
  if(what == "h" || what == "H") iwhat <- 2
  if(nocov == 0) yy <- xhare(iwhat, xx, yy, fit)
  else yy <- xhare(iwhat, xx, fit = fit)
  if(what == "s" || what == "S") yy <- 1 - yy
  if(missing(xlab)) xlab <- ""
  if(missing(ylab)) ylab <- ""
  if(missing(type)) type <- "l"
  xx <- (0:(n - 1))/(n - 1) * (u4 - u3) + u3
  if(!add) plot(xx, yy, xlab = xlab, ylab = ylab, type = type, ...)
  else lines(xx, yy, type = type, ...)
}
hare.summary <- function(fit)
{
  s3 <- as.vector(t(fit$logl))
  s3[is.na(s3)] <- 0
  s1 <- as.vector(t(fit$fcts))
  s2 <- as.vector(fit$knots)
  s1[is.na(s1)] <- -1
  s2[is.na(s2)] <- -1
  .C("ssumm",
     as.double(fit$penalty),
     as.integer(fit$sample),
     as.double(s3),
     as.integer(length(s3)/2),
     as.double(s2),
     as.double(s1),
     as.integer(fit$ndim),
     as.integer(fit$ncov))
  invisible()
}
dhare <- function(...)
{
  xhare(3, ...)
}
hhare <- function(...)
{
  xhare(2, ...)
}
phare <- function(...)
{
  xhare(0, ...)
}
qhare <- function(...)
{
  xhare(1, ...)
}
rhare <- function(n, ...)
{
  xhare(1, runif(n), ...)
}
xhare <- function(iwhat, q, cov, fit)
{
  # mainly messing with the covariates 
  if(missing(fit)) {
    fit <- cov
    if(is.null(fit$ncov)) stop("fit missing")
  }
  if(fit$ncov == 0) {
    if(!missing(cov) && !missing(fit))
      stop("there should be no covariates")
    else cov <- 0
  }
  else {
    if(is.matrix(cov) == F) cov <- matrix(cov, ncol = fit$ncov)
    nd <- length(cov[, 1])
    nc <- length(cov[1,  ])
    nq <- length(q)
    if(nc != fit$ncov) stop("not the right number of covariates")
    if(nd != 1 && nq != 1 && nd != nq) stop("no matching number of cases")
    if(nq == 1) q <- rep(q, nd)
    if(nd == 1 && nq != 1)
      cov <- matrix(cov, nrow = nq, ncol = nc, byrow = T)
  }
  fit$fcts <- as.vector(t(fit$fcts))
  fit$fcts[is.na(fit$fcts)] <- -1
  fit$knots <- as.vector(fit$knots)
  fit$knots[is.na(fit$knots)] <- 0
  z <- .C("sphare",
          as.integer(fit$ncov),
          as.integer(fit$ndim),
          as.integer(length(q)),
          as.double(cov),
          as.integer(iwhat),
          q = as.double(q),
          as.double(fit$knots),
          as.double(fit$fcts))
  z$q
}

# For Windows. B.D. Ripley 97/12/31
.First.lib <- function(library, section) 
{
   if(version$major == 3) {
     if(version$minor < 2) stop("This library needs S-Plus 3.2 or later")
     else sec <- "sp3"
   }
   if(version$major == 4) sec <- "sp4"
   dyn.load(paste(library, section, sec, c("mpack.obj", "hareall.obj"), sep="/"))
}
